# Copyright 2014, 2015, 2016, 2017 Kevin Ryde
#
# This file is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation; either version 3, or (at your option) any later
# version.
#
# This file is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU General Public License along
# with this file.  See the file COPYING.  If not, see
# <http://www.gnu.org/licenses/>.

package MyCCurve;
use strict;
use Carp 'croak';
use base 'Exporter';
use List::Util 'max','min';
use Math::Matrix;
use Memoize;

# uncomment this to run the ### lines
# use Smart::Comments;

our @EXPORT_OK = ('to_bigrat',
                  'CountLowZeros',
                  'dir','turn','turn_at_xy',
                  'R','L','L_formula',
                  'hull_expand',
                  'points_to_centroid',
                  'points_str','points_str_pari','points_to_bigrat',
                  'mul','div',
                  'xy_rotate_plus90','xy_rotate_minus90',
                  'xy_rotate_minus45','transform_to_horizontal',
                  'xy_to_dir4','xy_to_dir8',
                  'rect_corners_sort',
                  'is_integer','is_pow2','chamfer',
                  'G_xy',
                  'PartN','PartSegments','PartEnd',
                  'TopMiddlePre','TopMiddlePost',
                  '$Rpred_state_aref','Rpred_k_flat','NonRpred_k_flat',
                  'RpredInf_flat','NonRpredInf_flat',
                  '$Lpred_state_table_aref','Lpred_k_flat','NonLpred_k_flat',
                  'LpredInf_flat','NonLpredInf_flat',
                  'xyxy_to_fifteen',
                  'xydxdy_reverse',
                  'xydxdy_triangle_neighbour_start',
                  'xydxdy_triangle_neighbour_end',
                  'xydxdy_triangle_neighbours',
                  'xyxy_xyxy_have_overlap','xyxy_convex_hull_overlap',
                  'xydxdy_triangle_to_gpc',
                  '@dir4_to_dx','@dir4_to_dy',
                 );
our %EXPORT_TAGS = (all => \@EXPORT_OK);


sub to_bigrat {
  my ($n) = @_;
  require Math::BigRat;
  return Math::BigRat->new($n);
}

our @dir4_to_dx = (1,0,-1,0);
our @dir4_to_dy = (0,1,0,-1);

#------------------------------------------------------------------------------
# dir() and turn()

sub dir {
  my ($n) = @_;
  my $dir = 0;
  for ( ; $n; $n>>=1) {
    $dir += ($n & 1);
  }
  return $dir;
}

# count trailing 0s
sub turn {
  my ($n) = @_;
  my $turn = 1;
  for ( ; ($n&1)==0; $n>>=1) {
    $turn--;
    unless ($n) { croak "turn() is for n>=1"; }
  }
  return $turn;
}

# count trailing 0s
sub turn_at_xy {
  my ($x,$y) = @_;
  my $turn = 1;
  until (xy_mod_b($x,$y)) {  # until odd z
    $turn--;
    unless ($x || $y) { croak "turn() is for x,y != 0,0"; }
    ($x,$y) = div($x,$y);
    $x /= 2; $y /=2;
  }
  return $turn;
}

sub xy_mod_b {
  my ($x,$y) = @_;
  return ($x+$y) & 1;
}

#------------------------------------------------------------------------------
# Right Boundary

# Right boundary length
# R[k] = /  7*2^h - 2k - 6     if k even
#        \ 10*2^h - 2k - 6     if k odd
#        where h = floor(k/2)
sub R {
  my ($k) = @_;
  my $h = int($k/2);
  return ($k % 2 ? 10 : 7)*2**$h - 2*$k - 6;
}

sub L_formula {
  my ($k) = @_;
  return ($k&1 ? 78 : 55)*(1 << (($k>>1)-2)) + 14*$k - 130;
}
sub L {
  my ($k) = @_;
  return ($k <= 5 ? (1<<$k) : L_formula($k));
}


#------------------------------------------------------------------------------
# Centroid

sub G_imag_frac {
  my ($k) = @_;
  return 1/2*(1 - 1/2**$k);
}
sub G_xy {
  my ($k) = @_;
  require Math::Complex;
  my $bk = Math::Complex->make(1,1) ** $k;
  my $z = $bk + $bk * G_imag_frac($k) * Math::Complex->make(0,-1);
  return ($z->Re, $z->Im);
}

#------------------------------------------------------------------------------

# return $x,$y multiplied by 1+i
sub mul {
  my ($x,$y) = @_;
  # (x,y) + rot+90(x,y), is multiply (i+1)
  return ($x - $y,
          $y + $x);
}

# return $x,$y multiplied by 1-i
sub div {
  my ($x,$y) = @_;
  # (x,y) + rot+90(x,y), is multiply (1-i)
  return ($x + $y,
          $y - $x);
}

# return $x,$y unfolded around pivot $ex,$ey
sub unfold {
  my ($x,$y, $ex,$ey) = @_;
  ($x,$y) = xy_rotate_plus90($x,$y);  # rotate +90
  return ($ex+$x, $ey+$y);
}

sub xy_rotate_plus90 {
  my ($x,$y) = @_;
  return (-$y,$x);  # rotate +90
}
sub xy_rotate_minus90 {
  my ($x,$y) = @_;
  return ($y,-$x);  # rotate -90
}

sub xy_rotate_minus45 {
  my ($x,$y) = @_;
  return (($x+$y)/sqrt(2), ($y-$x)/sqrt(2));  # rotate -45
}

sub transform_to_horizontal {
  my ($x,$y, $level) = @_;
  while ($level >= 2) {
    ($x,$y) = xy_rotate_minus90($x,$y);
    $level -= 2;
  }
  if ($level) {
    ($x,$y) = div($x,$y);
  }
  return (-$x,-$y);
}

# return 0..4
sub xy_to_dir4 {
  my ($x, $y) = @_;
  require Math::NumSeq::PlanePathDelta;
  return Math::NumSeq::PlanePathDelta::_dxdy_to_dir4($x,$y);
}
# return 0..7
sub xy_to_dir8 {
  my ($x, $y) = @_;
  return 2*xy_to_dir4($x,$y);
}


sub rect_corners_sort {
  my ($x1,$y1, $x2,$y2) = @_;
  if ($x1 > $x2) {
    ($x1,$x2) = ($x2,$x1);
  }
  if ($y1 > $y2) {
    ($y1,$y2) = ($y2,$y1);
  }
  return ($x1,$y1, $x2,$y2);
}



#------------------------------------------------------------------------------

sub PartN {
  my ($k) = @_;
  return (2**$k + ($k%2==0 ? -1 : 1))/3;
}
sub PartSegments {
  my ($k) = @_;
  return PartN($k+1) - PartN($k);
}
# sub PartEnd {
#   my ($k) = @_;
#   return PartStart($k+1);
# }
# CHECK {
#   PartEnd(4) == 11 or die;
#   my @PartEnd_samples = (1,1,3,5,11,21,43);
#   foreach my $k (0 .. $#PartEnd_samples) {
#     # print PartEnd($k),",";
#     PartEnd($k) == $PartEnd_samples[$k]
#       or die "k=$k  got ",PartEnd($k)," want ",$PartEnd_samples[$k];
#   }
# }

sub TopMiddlePre {
  my ($k) = @_;
  if ($k <= 1) { return 0; }
  return (5*2**$k - ($k%2==0 ? 8 : 4))/12;
}
sub TopMiddlePost {
  my ($k) = @_;
  if ($k <= 1) { return $k; }
  return (7*2**$k - ($k%2==0 ? 4 : 8))/12;
}

our $Rpred_state_aref = [ [1,2],   # r0
                          [1,3],
                          [4,2],   # r2
                          [4,5],
                          [6,3],   # r4
                          [7,8],
                          [8,7],   # r6
                          [6,5],
                          [8,8],   # 8=non
                        ];
use constant::defer NonRpred_k_flat => sub {
  require MyFLAT;
  MyFLAT::aref_to_FLAT_DFA($Rpred_state_aref, name => 'NonRpred_k');
};
use constant::defer Rpred_k_flat => sub {
  my $f = NonRpred_k_flat()->complement;
  $f->{'name'} = 'Rpred_k';
  return $f;
};

use constant::defer NonRpredInf_flat => sub {
  require MyFLAT;
  MyFLAT::aref_to_FLAT_DFA($Rpred_state_aref,
                           starting => 1,  # as if high run 0s
                           name => 'NonRpredInf');
};
use constant::defer RpredInf_flat => sub {
  my $f = NonRpredInf_flat()->complement;
  $f->{'name'} = 'RpredInf';
  return $f;
};


#------------------------------------------------------------------------------
# Lpred

our $Lpred_state_table_aref =
  [ [ 1,  2],
    [ 1,  3],
    [ 4,  2],
    [ 5,  6],
    [ 7,  8],
    [ 7,  9],
    [10, 11],
    [12, 13],
    [14,  6],
    [14, 15],
    [16, 17],
    [18, 19],
    [19, 20],
    [21, 22],
    [23,  9],
    [24, 25],
    [12, 26],
    [47, 46],
    [45, 27],
    [41, 40],
    [28, 46],
    [45, 47],
    [29, 11],
    [30, 31],
    [16, 32],
    [48, 34],
    [33, 48],
    [34, 42],
    [43, 35],
    [48, 36],
    [35, 48],
    [37, 22],
    [47, 48],
    [45, 49],
    [49, 38],
    [39, 49],
    [49, 46],
    [48, 47],
    [49, 39],
    [38, 49],
    [41, 42],
    [43, 40],
    [44, 49],
    [49, 44],
    [43, 42],
    [49, 48],
    [48, 49],
    [48, 48],
    [49, 49],
    [49, 49],
  ];
our $Lpred_non = 49;
use constant::defer NonLpred_k_flat => sub {
  require MyFLAT;
  MyFLAT::aref_to_FLAT_DFA($Lpred_state_table_aref,
                           name => 'NonLpred_k',
                           accepting => $Lpred_non);
};
use constant::defer Lpred_k_flat => sub {
  my $f = NonLpred_k_flat()->complement;
  $f->{'name'} = 'Lpred_k';
  return $f;
};

use constant::defer NonLpredInf_flat => sub {
  require MyFLAT;
  MyFLAT::aref_to_FLAT_DFA($Lpred_state_table_aref,
                           starting => 1,  # as if high run 0s
                           name => 'NonLpredInf');
};
use constant::defer LpredInf_flat => sub {
  my $f = NonLpredInf_flat()->complement;
  $f->{'name'} = 'LpredInf';
  return $f;
};


#------------------------------------------------------------------------------
# convex hull

sub hull_points_and_end {
  my ($k) = @_;
  my $points = [ [0,0], [1,0], ];
  my ($ex,$ey) = (1,0);
  foreach my $k (1 .. $k) {
    ($points,$ex,$ey) = hull_expand($points,$ex,$ey);
  }
  return ($points,$ex,$ey);
}
sub hull_expand {
  my ($points, $ex,$ey, %options) = @_;
  ### hull_expand() ...

  if (@$points < 5) {
    ### small ...
    $points = [ points_midpoint($points->[0],$points->[1], 1,4),
                points_midpoint($points->[0],$points->[1], 1,2),
                points_midpoint($points->[0],$points->[1], 3,4),
                @$points ];
  }

  push @$points, map{[ unfold(@$_, $ex,$ey) ]} @$points;
  ### at: "pxy=$ex,$ey  ".points_str($points)

  require Math::Geometry::Planar;
  my $planar = Math::Geometry::Planar->new;
  $planar->points($points);
  if (@$points >= 5) {
    ### convexhull2: points_str($points)
    $planar = $planar->convexhull2;
    $points = $planar->points;
  }
  unless ($options{'unsorted'}) {
    $points = hull_points_sort_by_angle($points, $ex,$ey);
  }
  ($ex,$ey) = mul($ex,$ey);
  return ($points, $ex,$ey);
}

sub points_midpoint {
  my ($p1,$p2, $num,$den) = @_;
  $num //= 1;
  $den //= 2;
  return [ $p1->[0]*$num/$den + $p2->[0]*($den-$num)/$den,
           $p1->[1]*$num/$den + $p2->[1]*($den-$num)/$den ];
}

sub hull_points_sort_by_angle {
  my ($points, $ex,$ey) = @_;
  return [ sort {hull_point_cmp_by_angle($a,$b, $ex,$ey)} @$points ];
}
sub hull_point_cmp_by_angle {
  my ($a,$b, $ex,$ey) = @_;
  require Math::NumSeq::PlanePathDelta;
  my $a4 = Math::NumSeq::PlanePathDelta::_dxdy_to_dir4(@$a);
  my $b4 = Math::NumSeq::PlanePathDelta::_dxdy_to_dir4(@$b);
  if ($a->[0] == 0 && $a->[1] == 0) { $a4 = 2.5; }
  if ($b->[0] == 0 && $b->[1] == 0) { $b4 = 2.5; }
  my $o4 = Math::NumSeq::PlanePathDelta::_dxdy_to_dir4($ex,$ey);
  $a4 = ($a4 - $o4 + -1.2 - 4); while ($a4 < 0) { $a4 += 4; }
  $b4 = ($b4 - $o4 + -1.2 - 4); while ($b4 < 0) { $b4 += 4; }
  return ($a4 <=> $b4);
}


#------------------------------------------------------------------------------
# sequences

sub bit_above_lowest_one {
  my ($n) = @_;
  if ($n < 1) {
    croak "bit_above_lowest_one() is for n>=1";
  }
  for (;;) {
    if (! $n || ($n % 2) != 0) {
      last;
    }
    $n = int($n/2);
  }
  $n = int($n/2);
  return ($n % 2);
}
sub bit_above_lowest_zero {
  my ($n) = @_;
  for (;;) {
    if (($n % 2) == 0) {
      last;
    }
    $n = int($n/2);
  }
  $n = int($n/2);
  return ($n % 2);
}

sub CountLowZeros {
  my ($n) = @_;
  if ($n == 0) { die; }
  my $count = 0;
  until ($n % 2) {
    $count++;
    $n /= 2;
  }
  return $count;
}


#------------------------------------------------------------------------------
# Math::Geometry::Planar things

sub points_to_bigrat {
  my ($points) = @_;
  return [ map { [ to_bigrat($_->[0]), to_bigrat($_->[1]) ] } @$points ];
}

sub points_to_centroid {
  my ($points) = @_;
  if (@$points < 3) {
    return [ ($points->[0]->[0] + $points->[1]->[0]) / 2,
             ($points->[0]->[1] + $points->[1]->[1]) / 2 ];
  }
  require Math::Geometry::Planar;
  my $planar = Math::Geometry::Planar->new;
  $planar->points($points);
  return $planar->centroid;
}

# $points is an arrayref of [$x,$y] pairs.
# Return a string "x,y x,y ...".
sub points_str {
  my ($points) = @_;
  return  join(' ',map{join(',',@$_)}@$points)
}

# $points is an arrayref of [$x,$y] pairs.
# Return a string "x,y x,y ...".
sub points_str_pari {
  my ($points) = @_;
  return  join(', ',map{point_str_pari($_)}@$points)
}
sub point_str_pari {
  my ($p) = @_;
  my ($x,$y) = @$p;
  my $ret = '';
  if ($x != 0 || $y == 0) {
    $ret = "$x";
  }
  if ($y) {
    if ($ret ne '' && $y > 0) {
      $ret .= '+';
    }
    if ($y == 1) {
      $ret .= 'I';
    } elsif ($y == -1) {
      $ret .= '-I';
    } else {
      $ret .= "$y*I";
    }
  }
  return $ret;
}

#------------------------------------------------------------------------------
sub is_integer {
  my ($n) = @_;
  return $n == int($n);
}
sub is_square {
  my ($n) = @_;
  my $s = int(sqrt($n));
  my $ss = $s*$s;
  if ($ss > $n) {
    $s--;
    $ss = $s*$s;
  }
  return $ss == $n;
}

sub is_pow2 {
  my ($n) = @_;
  (($n >> 1) < $n) or die "is_pow2() infinite";

  while ($n > 1) {
    if ($n & 1) {
      return 0;
    }
    $n >>= 1;
  }
  return ($n == 1);
}

sub chamfer {
  my ($delta, $name) = @_;
  $name //= '\MyChamfer';
  ### $delta
  if ($delta == 0) { return ''; }
  if ($delta == 1) { return "+$name"; }
  if ($delta == -1) { return "-$name"; }
  return sprintf '%s%s*%s', ($delta>=0?'+':''), $delta, $name;
}

#------------------------------------------------------------------------------

# with start of segment 1 as the origin
my @fifteen_to_xyxy = (undef,
                       [ 0,0, -1,0 ],   # 1
                       [ -1,0, 0,0 ],   # 2
                       [ -1,-1, -1,0 ], # 3
                       [ -1,0, -1,-1 ], # 4
                       [ -2,0, -1,0 ],  # 5
                       [ -1,0, -2,0 ],  # 6
                       [ -1,1, -1,0 ],  # 7
                       [ -1,0, -1,1 ],  # 8
                       [ -1,1, 0,1 ],   # 9
                       [ 0,1, 0,0 ],    # 10
                       [ 0,0, 0,1 ],    # 11
                       [ 1,0, 0,0 ],    # 12
                       [ 0,0, 1,0 ],    # 13
                       [ 0,-1, 0,0 ],   # 14
                       [ 0,0, 0,-1 ],   # 15
                      );
my %xyxy_to_fifteen;
foreach my $i (1 .. $#fifteen_to_xyxy) {
  my ($x1,$y1, $x2,$y2) = @{$fifteen_to_xyxy[$i]};
  $xyxy_to_fifteen{"$x1,$y1, $x2,$y2"} = $i;
}

sub xyxy_to_fifteen {
  my ($x1,$y1, $x2,$y2) = @_;
  return $xyxy_to_fifteen{"$x1,$y1, $x2,$y2"};
}

sub xyxy_fifteen_part {
  my ($x1,$y1, $x2,$y2, $i) = @_;
  my $dx = $x2 - $x1;
  my $dy = $y2 - $y1;

  my $f = $fifteen_to_xyxy[$i];
  my ($fx1,$fy1, $fx2,$fy2) = @$f;

  my $dir = xy_to_dir4($dx,$dy);
  foreach (-1 .. $dir) {
    ($x1,$y1) = xy_rotate_minus90($x1,$y1);
    ($x2,$y2) = xy_rotate_minus90($x2,$y2);
  }
  ($x2-$x1 == -1 && $y2-$y1 == 0) or die;

  ($x1,$y1, $x2,$y2) = ($x1+$fx1,$y1+$fy1, $x1+$fx2,$y1+$fy2);

  foreach (-1 .. $dir) {
    ($x1,$y1) = xy_rotate_plus90($x1,$y1);
    ($x2,$y2) = xy_rotate_plus90($x2,$y2);
  }
  return ($x1,$y1, $x2,$y2);
}


#   <------  -1,1  <----------  0,1  <------
#   ------>        ---------->       ------>
#             ^ |               ^ |
#             | |               | |
#             | v               | v
#   <------  -1,0  <==========  0,0  <------
#   ------>        ---------->       ------>
#             ^ |               ^ |
#             | |               | |
#             | v               | v
#
#            -1,1  <----------  0,1
#

sub polygons_any_intersection {
  my ($points1, $points2) = @_;
  foreach my $i (0 .. $#$points1) {
    foreach my $j (0 .. $#$points2) {
      # ### intersection: [ $points1->[$i-1], $points1->[$i], $points2->[$j-1], $points2->[$j] ]
      my $p1 = $points1->[$i-1];
      my $p2 = $points1->[$i];
      my $p3 = $points2->[$j-1];
      my $p4 = $points2->[$j];

      if (Math::Geometry::Planar::SegmentIntersection([$p1,$p2,$p3,$p4])) {
        return 1;
      }
      if (Math::Geometry::Planar::Colinear([$p1,$p2,$p3])
          && (Math::Geometry::Planar::DistanceToSegment([$p1,$p2, $p3]) == 0
              || Math::Geometry::Planar::DistanceToSegment([$p1,$p2, $p4]) == 0
              || Math::Geometry::Planar::DistanceToSegment([$p3,$p4, $p1]) == 0
              || Math::Geometry::Planar::DistanceToSegment([$p3,$p4, $p2]) == 0)) {
        return 1;
      }
    }
  }
  return 0;
}
use constant::defer xyxy_convex_hull_overlap => sub {
  ### xyxy_convex_hull_overlap() ...
  require Math::Geometry::Planar;
  my $eps = 1/64;
  my $points = [ [  0, 1-$eps ],  # P1
                 [ -1, 1-$eps ],  # P1'

                 [  0  + (1/2-$eps), 1/2 ],  # P2
                 [  -1 - (1/2-$eps), 1/2 ],  # P2'

                 [  0  + (1/2-$eps), 0 ],  # P3
                 [  -1 - (1/2-$eps), 0 ],  # P3'

                 [  0  + (1/4-$eps), -(1/4-$eps) ],  # P4
                 [  -1 - (1/4-$eps), -(1/4-$eps) ],  # P4'
               ];
  my $planar = Math::Geometry::Planar->new;
  $planar->points($points);
  $planar = $planar->convexhull2;
  $points = $planar->points;
  ### $points

  my %overlap;
  require Tie::IxHash;
  tie %overlap, 'Tie::IxHash';

  $overlap{'0,0, -1,0'} = 1;  # own hull

  foreach my $dir (0 .. 3) {
    foreach my $y (-5 .. 5) {
      foreach my $x (-5 .. 5) {

        my $points2 = [];
        foreach my $p (@$points) {
          my ($px,$py) = @$p;
          foreach (1 .. $dir) { ($px,$py) = xy_rotate_plus90($px,$py); }
          push @$points2, [$px + $x, $py + $y];
        }

        if (polygons_any_intersection($points, $points2)) {
          my $dx = -1;
          my $dy = 0;
          foreach (1 .. $dir) { ($dx,$dy) = xy_rotate_plus90($dx,$dy); }

          my $x2 = $x + $dx;
          my $y2 = $y + $dy;
          ### overlap: "dir=$dir  key $x,$y, $x2,$y2"
          $overlap{"$x,$y, $x2,$y2"} = 1;
        }
      }
    }
  }

  my $num_overlaps = scalar(keys %overlap);
  ### %overlap
  # print "total $num_overlaps overlapping convex hulls\n";
  \%overlap;
};
sub xyxy_xyxy_have_overlap {
  my ($x1,$y1, $x2,$y2, $x3,$y3, $x4,$y4) = @_;
  ### xyxy_xyxy_have_overlap: "$x1,$y1, $x2,$y2, $x3,$y3, $x4,$y4"

  $x2 -= $x1; $y2 -= $y1;
  $x3 -= $x1; $y3 -= $y1;
  $x4 -= $x1; $y4 -= $y1;
  $x1 = 0; $y1 = 0;

  until ($x2 == -1 && $y2 == 0) {
    ($x1,$y1) = xy_rotate_plus90($x1,$y1);
    ($x2,$y2) = xy_rotate_plus90($x2,$y2);
    ($x3,$y3) = xy_rotate_plus90($x3,$y3);
    ($x4,$y4) = xy_rotate_plus90($x4,$y4);
  }

  xyxy_convex_hull_overlap();
  ### overlap norm: "$x1,$y1, $x2,$y2, key "."$x3,$y3, $x4,$y4"
  return xyxy_convex_hull_overlap()->{"$x3,$y3, $x4,$y4"};



  # my $dx1 = $x2 - $x1;  my $dx3 = $x4 - $x3;
  # my $dy1 = $y2 - $y1;  my $dy3 = $y4 - $y3;
  # $x1 *= 4;  $x2 *= 4;  $x3 *= 4;  $x4 *= 4;
  # $y1 *= 4;  $y2 *= 4;  $y3 *= 4;  $y4 *= 4;
  #
  # ### expanded: "$x1,$y1, $x2,$y2,   $x3,$y3, $x4,$y4"
  #
  # # lengthen by 1/2 each end
  # $x1 -= 2*$dx1; $y1 -= 2*$dy1;  $x3 -= 2*$dx3; $y3 -= 2*$dy3;
  # $x2 += 2*$dx1; $y2 += 2*$dy1;  $x4 += 2*$dx3; $y4 += 2*$dy3;
  #
  # ($dx1,$dy1) = xy_rotate_minus90($dx1,$dy1);  # right side
  # ($dx3,$dy3) = xy_rotate_minus90($dx3,$dy3);  # right side
  #
  # # widen by 1 right and 1/4 left
  # $x1 -= 1*$dx1; $y1 -= 1*$dy1;  $x3 -= 1*$dx3;  $y3 -= 1*$dy3;  # 1/4 left
  # $x2 += 4*$dx1; $y2 += 4*$dy1;  $x4 += 4*$dx3;  $y4 += 4*$dy3;  # 1 right
  #
  # # now boxes which are convex hull (exclusive)
  # ### boxes: "$x1,$y1, $x2,$y2,   $x3,$y3, $x4,$y4"
  #
  # if (max($x1,$x2) <= min($x3,$x4)
  #     || max($x3,$x4) <= min($x1,$x2)) {
  #   ### no overlap X ...
  #   return 0;
  # }
  # ### X overlap ...
  # if (max($y1,$y2) <= min($y3,$y4)
  #     || max($y3,$y4) <= min($y1,$y2)) {
  #   ### no overlap Y ...
  #   return 0;
  # }
  # ### Y overlap too ...
  # return 1;




  # # locations $x1,$y1, $x2,$y2 expands into
  # my %seen;
  # foreach my $i (1 .. 15) {
  #   my ($px1,$py1, $px2,$py2) = xyxy_fifteen_part($x1,$y1, $x2,$y2, $i);
  #   ### seen: "xy12  $i  $px1,$py1, $px2,$py2"
  #   $seen{"$px1,$py1, $px2,$py2"} = $i;
  # }
  #
  # # see if $x3,$y3, $x4,$y4 expands into any of them too
  # foreach my $j (1 .. 15) {
  #   my ($px1,$py1, $px2,$py2) = xyxy_fifteen_part($x3,$y3, $x4,$y4, $j);
  #   my $i = $seen{"$px1,$py1, $px2,$py2"};
  #   ### match: "xy34  $j  $px1,$py1, $px2,$py2  is ".($i//'[none]')
  #   if ($i) {
  #     return "$i and $j";
  #   }
  # }
  # return 0;
}


#------------------------------------------------------------------------------

sub xydxdy_reverse {
  my ($x,$y, $dx,$dy) = @_;
  return ($x+$dx,$y+$dy, -$dx,-$dy);
}

#  ----->
# ^  v
# |>
# |
sub xydxdy_triangle_neighbour_start {
  my ($x,$y, $dx,$dy) = @_;
  ($dx,$dy) = xy_rotate_plus90($dx,$dy);
  return ($x-$dx, $y-$dy, $dx,$dy);
}

#  ----->
#    v   |
#       <|
#        v
sub xydxdy_triangle_neighbour_end {
  my ($x,$y, $dx,$dy) = @_;
  return ($x+$dx, $y+$dy, xy_rotate_minus90($dx,$dy));
}

# x,y,dx,dy is a triangle.  Return the 3 triangles which have sides in
# common with this triangle, as 3 arrayrefs [x,y,dx,dy].
#
#       x,y    A    x2,y2
#           <-----
#        * --------> *
#        ^ \       / |
#       B|   mx,my   |C
#        |           v
#
sub xydxdy_triangle_neighbours {
  my ($x,$y, $dx,$dy) = @_;
  return ([xydxdy_reverse($x,$y,$dx,$dy)],
          [xydxdy_triangle_neighbour_start($x,$y,$dx,$dy)],
          [xydxdy_triangle_neighbour_end($x,$y,$dx,$dy)]);
}

sub xydxdy_triangle_to_gpc {
  my ($x,$y, $dx,$dy) = @_;
  my ($rx,$ry) = xy_rotate_minus90($dx,$dy);
  my $x2 = $x + $dx;
  my $y2 = $y + $dy;
  my $mx = $x + $dx/2 + $rx/2;
  my $my = $y + $dy/2 + $ry/2;

  my $planar = Math::Geometry::Planar->new;
  $planar->points([ [$x,$y], [$mx,$my], [$x2,$y2] ]);
  return $planar->convert2gpc;
}


#------------------------------------------------------------------------------
1;
__END__
